/*

  PX - File compressor

  Based on PAQ1 by Matt Mahoney <mmahoney@cs.fit.edu>

  Written by Ilia Muraviev <encode@narod.ru>

*/
#include <cstdio>
#include <cmath>
//using namespace std;

static FILE *infile; // input file stream
static FILE *outfile; // output file stream

enum TMode {mCompress, mDecompress};

static const int PSCALE = 4096;

class TCounter {
public:
  unsigned short checksum;
  unsigned char n0;
  unsigned char n1;

  TCounter(int c = 0): checksum(c), n0(0), n1(0) {}

  inline int Priority() {return (n0 + n1);}

  inline void Add(int y) {
    if (y) {
      if (n1 < 255) n1++;
      if (n0 > 2) n0 -= n0 >> 1; // n0 -= n0 / 2
    }
    else {
      if (n0 < 255) n0++;
      if (n1 > 2) n1 -= n1 >> 1; // n1 -= n1 / 2
    }
  }
};

class THashTable {
private:
  enum {N = 24, M = ((1 << N) - 1)};

  TCounter *counter;

public:
  THashTable(): counter(new TCounter[(1 << N) + 1]) {}

  ~THashTable() {delete[] counter;}

  inline TCounter &operator[](unsigned int i) {
    const unsigned int checksum = i >> 16;

    const unsigned int lo = i & M;
    const unsigned int hi = lo + 2;

    int bj;
    int bp = 0x1ff;

    for (unsigned int j = lo; j < hi; j++) {
      TCounter &c = counter[j];

      if (checksum == c.checksum) return (c);

      const unsigned int p = c.Priority();
      if (p == 0) return (c = TCounter(checksum));

      if (p < bp) {
        bp = p;
        bj = j;
      }
    }

    return (counter[bj] = TCounter(checksum));
  }
};

class TModel {
private:
  enum {N = 5, SSE2 = 32, SSESCALE = (1024 / SSE2)};

  class TSSEMap {
  private:
    unsigned short table[PSCALE];

  public:
    TSSEMap() {
      for (int i = 0; i < PSCALE; i++) {
        int p = static_cast<int>((64 * log((i + 0.5f) / ((PSCALE - 0.5f) - i)))
            + 512);

        if (p > 1023) p = 1023;
        else if (p < 0) p = 0;

        table[i] = p;
      }
    }

    inline int operator()(int p) {return table[p];}
  };

  class TSSEContext {
  public:
    unsigned char n;
    unsigned char t;

    TSSEContext(): n(0), t(0) {}

    inline int P() {return ((PSCALE * ((n * 64) + 1)) / ((t * 64) + 2));}

    inline void Update(int y) {
      if (y) n++;

      if (++t > 254) {
        n >>= 1; // n /= 2
        t >>= 1; // t /= 2
      }
    }
  };

  TSSEMap sseMap;
  TSSEContext sse[SSE2 + 1];

  TCounter counter0[256];
  TCounter counter1[65536];
  THashTable hashTable;

  int c0;
  int c1;
  int cn;

  unsigned int hash[10];
  TCounter *cp[10];

  unsigned int ssep;

public:
  TModel(): c0(1), c1(0), cn(1) {
    for (int k = 0; k < N; k++) {
      hash[k] = 0;
      cp[k] = &counter0[0];
    }

    int r = SSE2 + 1;

    for (int i = PSCALE - 1; i >= 0; i--) {
      const int p = (sseMap((i * PSCALE) / PSCALE) + (SSESCALE / 2))
          / SSESCALE;

      int t = ((PSCALE * PSCALE) / ((i + 1) * (PSCALE - i))) + 1;
      if (t > 254) t = 254;

      const int n = ((i * t) + (PSCALE / 2)) / PSCALE;

      for (int j = r - 1; j >= p; j--) {
        sse[j].n = n;
        sse[j].t = t;
      }

      r = p;
    }
  }

  inline int Predict() {
    int n0 = 1;
    int n1 = 1;

    for (int i = 0; i < N; i++) {
      n0 += cp[i]->n0 << i;
      n1 += cp[i]->n1 << i;
    }

    int p = (PSCALE * n1) / (n0 + n1);


    ssep = sseMap(p);

    const unsigned int w = ssep % SSESCALE;
    const unsigned int z = ssep >> 5; // ssep / SSESCALE

    p = (((((sse[z].P() * (SSESCALE - w)) + (sse[z + 1].P() * w))
        >> 5) // (...) / SSESCALE
        * 3) + p)
        >> 2; // (...) / 4

    return (p);
  }

  inline void Update(int y) {
    for (int i = 0; i < N; i++) cp[i]->Add(y);

    if ((cn += cn + y) >= 53) cn -= 53;
    c0 += c0 + y;

    if (c0 >= 256) {
      for (int i = N - 1; i > 0; i--)
        hash[i] = (hash[i - 1] + c0) * 987660757;

      c1 = c0 - 256;
      c0 = 1;
      cn = 1;
    }

    cp[0] = &counter0[c0];
    cp[1] = &counter1[(c1 << 8) + c0];

    const unsigned int x = (c0 << 24) + cn;

    for (int j = 2; j < N; j++)
      cp[j] = &hashTable[hash[j] + x];


    const unsigned int z = ssep >> 5; // ssep / SSESCALE

    sse[z].Update(y);
    sse[z + 1].Update(y);
  }
};

class TCodec {
private:
  TModel model;

  const TMode mode;
  FILE *f;

  unsigned int x1;
  unsigned int x2;
  unsigned int x;

  inline void EncodeBit(int y) {
    const unsigned int p = model.Predict();
    const unsigned int xdiff = x2 - x1;
    unsigned int xmid = x1;

    if (xdiff >= 0x4000000) xmid += (xdiff >> 12) * p;
    else if (xdiff >= 0x100000) xmid += ((xdiff >> 6) * p) >> 6;
    else xmid += (xdiff * p) >> 12;

    if (y)
      x2 = xmid;
    else
      x1 = xmid + 1;

    model.Update(y);

    while (((x1 ^ x2) & 0xff000000) == 0) {
      fputc(x2 >> 24, f);
      x1 <<= 8;
      x2 = (x2 << 8) + 0xff;
    }
  }

  inline int DecodeBit() {
    const unsigned int p = model.Predict();
    const unsigned int xdiff = x2 - x1;
    unsigned int xmid = x1;

    if (xdiff >= 0x4000000) xmid += (xdiff >> 12) * p;
    else if (xdiff >= 0x100000) xmid += ((xdiff >> 6) * p) >> 6;
    else xmid += (xdiff * p) >> 12;

    int y = 0;
    if (x <= xmid) {
      y = 1;
      x2 = xmid;
    }
    else
      x1 = xmid + 1;

    model.Update(y);

    while (((x1 ^ x2) & 0xff000000) == 0) {
      x1 <<= 8;
      x2 = (x2 << 8) + 0xff;

      int c = fgetc(f);
      if (c == EOF) c = 0;

      x = (x << 8) + c;
    }
    return (y);
  }

public:
  TCodec(TMode m, FILE *p): mode(m), f(p), x1(0), x2(0xffffffff), x(0) {
    if (mode == mDecompress) {
      for (int i = 0; i < 4; i++) {
        int c = fgetc(f);
        if (c == EOF) c = 0;
        x = (x << 8) + c;
      }
    }
  }

  ~TCodec() {
    if (mode == mCompress) {
      while (((x1 ^ x2) & 0xff000000) == 0) {
        fputc(x2 >> 24, f);
        x1 <<= 8;
        x2 = (x2 << 8) + 0xff;
      }
      fputc(x2 >> 24, f);
    }
  }

  inline void Encode(unsigned char c) {
    for (int i = 7; i >= 0; i--)
      EncodeBit((c >> i) & 1);
  }

  inline unsigned char Decode() {
    unsigned char c = 0;

    for (int i = 0; i < 8; i++)
      c += c + DecodeBit();

    return (c);
  }
};

static const int N = (1 << 16); // Buffer size: 64 KB
static const int X = (1 << 16);
static const int F = 181; // Match flag

static unsigned char buf[N]; // Data buffer
static unsigned short tab[X]; // Index table

static int i;
static int n;

static int encode() {
  fseek(infile, 0, SEEK_END);
  int size = ftell(infile);
  fseek(infile, 0, SEEK_SET);
  if (size < 0) {
    fprintf(stderr, "file too large\n");
    return (1);
  }

  fwrite(&size, 1, sizeof(size), outfile);

  TCodec codec(mCompress, outfile);

  while ((n = fread(&buf, 1, N, infile)) > 0) {
    for (i = 0; i < X; i++) // Initialize table
      tab[i] = 0;

    codec.Encode(buf[0]);

    if (n > 1)
      codec.Encode(buf[1]);


    i = 2;
    while (i < n) {
      const unsigned short c = *(reinterpret_cast<unsigned short *>(
        &buf[i - 2]));

      int p = tab[c]; // Match position
      tab[c] = i; // Update table

      int j = i;
      int l = 0; // Match length
      while ((buf[p] == buf[j]) && (l < 255) && (j < n)) {
        p++;
        j++;
        l++; // Increment match length
      }

      if (l < 8) { // Not long enough match
        codec.Encode(buf[i]); // Send a literal

        if (buf[i] == F) { // Special case (byte == match flag)
          codec.Encode(l); // Send match length
          if (l) i += l;
          else i++;
        }
        else i++;


      }
      else {
        codec.Encode(F); // Send match flag
        codec.Encode(l); // Send match length
        i += l;
      }

    }

  }
  return (0);
}

static void decode() {
  int size;
  fread(&size, 1, sizeof(size), infile);

  TCodec codec(mDecompress, infile);

  n = 0;

  while (n < size) {
    for (i = 0; i < X; i++) // Initialize table
      tab[i] = 0;

    buf[0] = codec.Decode();
    n++;

    if (n < size) {
      buf[1] = codec.Decode();
      n++;

      i = 2;
    }
    else
      i = 1;

    while ((i < N) && (n < size)) {
      buf[i] = codec.Decode();

      const unsigned short c = *(reinterpret_cast<unsigned short *>(
          &buf[i - 2]));

      int p = tab[c]; // Match position
      tab[c] = i; // Update table

      int l; // Match length
      if ((buf[i] == F) && ((l = codec.Decode()) > 0)) {
        n += l;

        while ((l > 0) && (i < N)) {
          buf[i] = buf[p];
          i++;
          p++;
          l--;
        }
      }
      else {
        i++;
        n++;
      }
    }
    fwrite(&buf, 1, i, outfile);
  }
}

static float ratio() {
 float a = static_cast<float>(ftell(infile));
 float b = static_cast<float>(ftell(outfile));

 if ((a > 0.0f) && (b > 0.0f)) {
  return ((b * 100.0f) / a);
 }
 return (0.0f);
}

int main(int argc, char **argv) {
 printf("px 1.0 (c) 2006 ilia muraviev\n");

 if ((argc != 4) || (argv[1][1] != 0) || (argv[1][0] != 'e')
   && (argv[1][0] != 'd')) {
  fprintf(stderr, "usage: px {e|d} <infile> <outfile>\n");
  return (1);
 }

 if (((infile = fopen(argv[2], "rb")) == NULL) || ((outfile
   = fopen(argv[3], "wb")) == NULL)) {
  fprintf(stderr, "i/o error\n");
  return (1);
 }

 if (argv[1][0] == 'e') {
  printf("encoding %s...\n", argv[2]);
  if (encode())
    return (1);
  printf("ratio: %.1f%%\n", ratio());
 } else {
  printf("decoding %s...\n", argv[2]);
  decode();
 }

 fclose(infile);
 fclose(outfile);

 printf("done\n");
 return (0);
}

